;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-
;;;
;;; last modified Time-stamp: <Jul-26-2006 03:23>

;;;; Copyright (c) 1992, 1993, 1994, 1995, 1996, 1997 by Paul Eric Nielsen.
;;;; The following software was developed by Paul Eric Nielsen for his
;;;; exclusive use.  This software, or excerpts thereof, may be freely
;;;; redistributed, in any medium, provided that you conspicuously and
;;;; appropriately publish on each copy an appropriate copyright notice.
;;;; It may not be sold for profit or incorporated in commercial documents
;;;; (e.g., published for sale on CD-ROM, floppy disks, books, magazines,
;;;; or other form) without the prior written permission of the copyright
;;;; holder.  This software is provided AS IS without any expressed or
;;;; implied warranty.

(in-package :cm-etf)

(defun fcomb (list n function)
  "Another version of the 'comb' function.  This one avoids consing by
  accepting an additional argument which is the function to be applied
  to each of the various combinations.
  The list passed to the function is destructively modified.  If you
  want to save it, you must copy it (avoiding the savings in cons cells).
  Example:
       (fcomb '(a b c) 2 <function>)
          ;=> will call <function> three times with the following arguments:
              (A B) (A C) (B C)
  "
  (let ((choice (make-list n)))
    ;; choices are items remaining to choose from
    ;; c-length is the length of choices
    ;; tail is the tail of choice that has yet to be filled in with valid
    ;;   values
    ;; t-length is the number of items to choose from choice, and should be
    ;;   = to the length of tail.
    (labels ((fcomb-1 (choices c-length tail t-length)
               (cond 
                ((null tail) (funcall function choice))
                ((>= c-length t-length)
                 (loop for l on choices
                       for i from c-length downto t-length
                       do (setf (car tail) (car l))
                       do (fcomb-1 (cdr l) (1- i) (cdr tail) (1- t-length))))))) 
      (fcomb-1 list (length list) choice n))))

(defun combinations (list n)
  "This passes a collector function to 'fcomb' which illustrates the
  use of 'fcomb' and yields exactly the same behavior as the other
  'comb' functions.
  Example:
       (combinations '(A B C) 2)
          ;=> ((B C) (A C) (A B))
  "
  (let ((choices nil))
    (fcomb list n #'(lambda (choice)
		       (push (copy-list choice) choices)))
    choices))

(defun cartesian (sets)
 "Returns the Cartesian product of a list of sets
  Example:
       (CARTESIAN '((A B) (C D) (E F)))
         ;=> ((B C F) (B C E) (B D F) (B D E) (A C F) (A C E) (A D F) (A D E))
  Non-recursive version."
 (let ((cartesian (list nil)))
   (dolist (set (reverse sets))
     (let ((new-cartesian nil))
       (dolist (elmt set)
         (dolist (cart cartesian)
           (push (cons elmt cart) new-cartesian)))
       (setf cartesian new-cartesian)))
   cartesian))
