;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-

;;; CM-ETF: music notation output in Finale format from Common Music
;;; Copyright (C) 2006 Niklas Kambeitz (niklas.kambeitz@mail.mcgill.ca)
;;;
;;; last modified Time-stamp: <07/05/21 17:00:58 nik>
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

;;  Sections:
;; -----------
;; SEQ->SCORES
;; SCORE->TREE
;; TREE->META
;; SCORE->META
;; META->ETF
;; SCORE->ETF
;; SEQ->ETF

;; TODO:
;; - fix (score->etf '((3 16/5 (60))) :time-sigs '((3 2)))
;; - fix tuplet numbers

(in-package :cm-etf)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;        SEQ->SCORES         ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun seq->events (seq)
  "e.g. (seq->events #!diatonic-layer-track-0)"
  (let ((result nil))
    (map-objects #'(lambda (midi-event)
                        (push (list (slot-value midi-event 'time)
                                    (slot-value midi-event 'duration)
                                    (round (keynum (slot-value midi-event 'keynum))))
                              result))
                    seq :class 'midi)
    (reverse result)))

(defun seq-rhythms (midi-sequence)
  "e.g. (seq-rhythms #!diatonic-layer-track-0)"
  (remove-duplicates
   (mapcar #'(lambda (x) (list (car x) (cadr x)))
           (seq->events midi-sequence))
   :test #'equal))

(defparameter *allowed-subdivisions* 
  '(8 12 10) 
  "subdivisions of the beat allowed for quantization")

(defun quantize2 (the-number &optional (subdivisions *allowed-subdivisions*))
  "e.g. (quantize2 21/20)"
  (caar (sort (mapcar #'(lambda (x)
                          (list x (abs (- x the-number))))
                      (mapcar #'(lambda (y)
                                  (quantize the-number (/ 1 y)))
                              subdivisions))
              #'< :key #'cadr)))

(defun closest-number (number list-of-numbers)
  "e.g. (closest-number 5 '(1 3 9))"
  (caar (sort (mapcar #'(lambda (x)
                          (list x (abs (- x number))))
                      list-of-numbers)
              #'< :key #'cadr)))

(defun quantize-score (score-list)
  "e.g. (quantize-score *score-list*)"
  (let ((list-of-times (mapcar #'quantize2 (find-all-events score-list))))
    (mapcar #'(lambda (x)
                (let ((closest-attack (closest-number (car x) list-of-times)))
                  (list closest-attack
                        (- (closest-number (+ (car x) (cadr x)) list-of-times) closest-attack)
                        (third x))))
            score-list)))

(defun seq->score (midi-sequence)
  "(seq->score #!diatonic-layer-track-0)"
  (let ((events-list (seq->events midi-sequence)))
    (quantize-score
     (mapcar #'(lambda (event)
                 (list (car event)
                       (cadr event)
                       (sort (mapcar #'third
                                     (remove-if-not #'(lambda (x)
                                                        (and (= (car event) (car x))
                                                             (= (cadr event) (cadr x))))
                                                    events-list))
                             #'<)))
             (seq-rhythms midi-sequence)))))

;;;==================================================================
;;; for multichannel sequences:

(defun used-channels (seq)
  "e.g.(used-channels *master-seq*)"
  (remove-duplicates 
   (mapcar #'(lambda (x) (slot-value x 'channel))
           (subobjects seq))))

(defun separate-channels (seq)
  "e.g. (separate-channels *master-seq*)"
  (let* ((result (mapcar #'list (used-channels seq))))
    (map-objects #'(lambda (midi-event)
                        (rplacd (last (nth (slot-value midi-event 'channel) result))
                                (list (list (slot-value midi-event 'time)
                                            (slot-value midi-event 'duration)
                                            (round (keynum (slot-value midi-event 'keynum)))))))
                    seq :class 'midi)
    (mapcar #'cdr result)))

(defun channel->score (channel-events)
  "e.g. (channel->score (car (separate-channels *master-seq*)))"
  (quantize-score
   (mapcar #'(lambda (event)
               (list (car event)
                     (cadr event)
                     (sort 
                      (mapcar #'(lambda (x)
                                  (round 
                                   (third x)))
                              (remove-if-not #'(lambda (x)
                                                 (and (= (car event) (car x))
                                                      (= (cadr event) (cadr x))))
                                             channel-events))
                      #'<)))
           channel-events)))

(defun seq->scores (seq)
  "e.g. (seq->scores *master-seq*)"
  (mapcar #'channel->score (separate-channels seq)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;        SCORE->TREE         ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; For a explanation of rhythm trees, see:
; http://recherche.ircam.fr/equipes/repmus/OpenMusic/Documentation/OMUserDocumentation/DocFiles/Reference/Rhythmtrees/Index.html

(defun measure-length (time-sig)
  "e.g. (measure-length '(4 4))"
  (* (car time-sig) (/ 4 (cadr time-sig))))

(defun sigs-length (time-sigs)
  "e.g. (sigs-length '((4 4) (3 4)))"
  (apply #'+ (mapcar #'measure-length time-sigs)))

(defun missing-length (score-length time-sigs)
  "e.g. (missing-length 10 '((4 4) (3 4)))"
  (- score-length (sigs-length time-sigs)))

(defun missing-bars (score-length time-sig)
  "e.g. (missing-bars 1 '(4 4))"
  (ceiling score-length (measure-length time-sig)))

(defun missing-sigs (score-length time-sig)
  "e.g. (missing-sigs 9 '(3 4))"
  (make-list (missing-bars score-length time-sig)
             :initial-element time-sig))

(defun fill-sigs (score-list time-sigs)
  "e.g. (fill-sigs '((0 3)) '((4 4) (4 4)))"
  (let ((total-length (score-length score-list))
        (sigs-length (sigs-length time-sigs)))
    (if (< sigs-length total-length)
        (append time-sigs 
                (missing-sigs (- total-length sigs-length)
                              (car (last time-sigs))))
        time-sigs)))

(defun score->durs (score-list)
  "e.g. (score->durs '((0 1) (3 2)))"
  (mapcar #'(lambda (x)
              (rationalize (/ x 4)))
  (om::x->dx (union '(0)
                    (find-all-events score-list)))))

(defun score->tree-raw (score-list &optional (time-sigs '(4 4)))
  "e.g. (score->tree-raw '((0 1) (3 2)) '(4 4))"
  (let ((time-sig-list (if (typep (car time-sigs) 'list)
                           time-sigs
                           (list time-sigs))))
     (om::simple->tree (score->durs score-list)
                       (fill-sigs score-list time-sig-list))))

;;;================================================================
;;; Tree Correction

;; OM::MAKE-PROPORTIONAL-CELL needs these helper functions 
;; to break up notationally impossible proportions, like 5.

(defun correct-replacement (value replacement) 
  "e.g. (correct-replacement 12 '(10 2))" 
  (cond ((minusp value) (mapcar #'- replacement)) 
        ((floatp value) (mapcar #'float replacement)) 
        (t (list (car replacement) (float (cadr replacement))))))

(defun replace-bad-value (value) 
  "e.g. (replace-bad-value -12)"
  (let ((abs-value (rationalize (abs value)))) 
    (if (member abs-value '(1 2 3 4 6 7 8 12 14 15 16 24 28 30 31 32 48 56 60 62)) 
      (list value) 
      (correct-replacement value (cadr (assoc abs-value '((5 (3 2)) 
                                                          (9 (6 3)) 
                                                          (10 (6 4)) 
                                                          (11 (8 3)) 
                                                          (13 (12 1)))))))))

(defun split-at-point (value point)
  "e.g. (split-at-point -12 11)"
  (if (< point (abs value))
    (correct-replacement value (list point (- (rationalize (abs value)) point)))
    (list value)))

(defun make-simple-tree (length)
  "e.g. (make-simple-tree 4)"
  (make-list (rationalize (abs length)) :initial-element 1))

(defun group-tree (tree)
  "e.g. (group-tree '(1 1 1))"
  (let ((first-value (car tree))
     ; (new-tree (copy-tree tree))
        result)
    (dotimes (i (ceiling (/ (length tree) 2)) (reverse result))
      (let ((x (nth (* i 2) tree))
            (y (nth (1+ (* i 2)) tree)))
        (cond ((and y (= x first-value) (= y first-value)) (push (+ x y) result))
              (t (progn (push x result) (if y (push y result)))))))))

(defun all-groupings (length-or-tree)
  "e.g. (all-groupings 9)"
  (cond ((and (listp length-or-tree)
              (equal (car length-or-tree) (group-tree (car length-or-tree))))
         length-or-tree)
        ((listp length-or-tree) (all-groupings (nconc (list (group-tree (car length-or-tree))) length-or-tree)))
        ((numberp length-or-tree) (all-groupings (list (make-simple-tree length-or-tree))))))

(defun find-breaker (now length divisions)
  "e.g. (find-breaker 3 5 (apply #'append '((0 8 9) (0 4 8 9) (0 2 4 6 8 9) (0 1 2 3 4 5 6 7 8 9))))"
  (find-if #'(lambda (x)
               (and (> x now) (< x (+ now length))))
           divisions))

(defun tree-p (obj) 
  (and (listp obj)
       (numberp (first obj))
       (not (null (second obj)))
       (= (length obj) 2)
       (listp (second obj))))

; (defun tree-p (obj) 
;   (and (listp obj)
;        (numberp (first obj))
;        (not (null (second obj)))
;        (listp (second obj))))

(defun leaf-extent (x)
  "e.g. (leaf-extent '(1 (2 1)))
        (leaf-extent -2.0)"
  (cond ((numberp x) (rationalize (abs x)))
        ((tree-p x) (car x))))

(defun note-list-sum (note-list)
  "e.g. (note-list-sum '(1 1 (1 (2 1)) 1))"
   (apply #'+ (mapcar #'leaf-extent note-list)))

(defun split-bad-values (leaf-list)
  "e.g. (split-bad-values '(4.0 9 3))"
  (if (tree-p leaf-list) leaf-list
      (let* ((total-length (note-list-sum leaf-list))
             (divisions (apply #'nconc 
                               (mapcar #'(lambda (x) 
                                           (cdr (om::dx->x 0 x)))
                                       (all-groupings total-length))))
             (now 0)
             (possible-integers '(32 16  8 4 2 1 ; undotted
                                  48 24 12 6 3   ; 1 dot
                                  56 28 14 7     ; 2 dots
                                  60 30 15       ; 3 dots
                                  62 31))        ; 4 dots
             result)
        (dolist (x leaf-list)
          (let ((abs-rat-x (cond ((numberp x) (abs (rationalize x)))
                                 ((listp x) (car x)))))
            (if (or (listp x)
                    (member abs-rat-x possible-integers))
                (push (list x) result)
                (push (split-at-point x (- (find-breaker now abs-rat-x divisions) now)) result))
            (setq now (+ now abs-rat-x))))
        (setf result (apply #'nconc (reverse result)))
        ;; recurse if there's still bad values:
        (if (every #'(lambda (x)
                       (or (listp x)
                           (member (abs (rationalize x)) possible-integers)))
                   result)
            result
            (split-bad-values result)))))

(defun split-measure (obj)
  "e.g. (cm-etf::split-measure '(1 (5 4)))
   e.g. (cm-etf::split-measure '(1 (1 (1 1)) (1 (1 1))))
   e.g. (cm-etf::split-measure '(1 (1 (1 1)) (1 (1.0 -1)) -1))"
  (cond ((numberp obj) obj)
        ((tree-p obj) (list (car obj)
                            (split-measure (split-bad-values (cadr obj)))))
        ((listp obj) (mapcar #'split-measure obj))))

(defun correct-tree (tree)
  "e.g. (cm-etf::correct-tree '(OM::? (((1 1) ((1 (5 4)))))))"
  (list (car tree)
        (mapcar #'(lambda (measure)
                    (list (car measure)
                          (split-measure (cadr measure))))
                (cadr tree))))

(defun score->tree (score-list &optional (time-sigs '(4 4)))
  "e.g. (score->tree '((0 1) (3 2)) '(4 4))"
  (let ((time-sig-list (if (typep (car time-sigs) 'list)
                           time-sigs
                           (list time-sigs))))
    (correct-tree
     (score->tree-raw score-list
                      (fill-sigs score-list 
                                 time-sig-list)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;        TREE->META          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(defun tree-p (obj) 
;  (and (listp obj)
;       (numberp (first obj))
;       (not (null (second obj)))
;       (listp (second obj))))

;(defun leaf-extent (x)
;  "e.g. (leaf-extent '(1 (2 1)))
;        (leaf-extent -2.0)"
;  (cond ((numberp x) (rationalize (abs x)))
;        ((tree-p x) (car x))))

;(defun note-list-sum (note-list)
;  "e.g. (note-list-sum '(1 1 (1 (2 1)) 1))"
;   (apply #'+ (mapcar #'leaf-extent note-list)))

(defun leaf->entry (leaf-value unit-value)
  "e.g. (leaf->entry 3 1/4)"
  ;; unit-value expressed in terms of whole notes. i.e. 1 = 1 whole note
  (list (duration->value
         (* (rationalize (abs leaf-value))
            unit-value
            4
            ))
        (if (floatp leaf-value) 'tied-to)
        nil                    ; tied-from placeholder (dealt with later)
        nil                    ; tuplet placeholder (dealt with later)
        nil                    ; break-beam placeholder (dealt with later)
        (if (minusp leaf-value) 'rest 'notes)))

; (dtrace::duntrace leaf->entry)

(defun round-duration (duration &optional (amount 1/16))
  "e.g. (round-duration 3)"
  (if (> amount duration) 
      (/ amount 2) 
      (round-duration duration (* amount 2))))

(defun process-tree (obj &optional denominator)
  "e.g. (process-tree '(1 (2 (1 (2 1)) 2)))"
  (cond ((numberp obj) (leaf->entry obj denominator))
        ((tree-p obj)
         (list (* (car obj)
                  (or denominator (car obj)))
               (mapcar #'(lambda (x)
                           (process-tree x ;leaf-value
                                         (* (car obj)
                                            (/ (round-duration (/ (/ (or denominator (car obj))
                                                                     (note-list-sum (cadr obj))
                                                                     )))))))
                       (cadr obj))))))

;(tree->meta '(? (((4 4) (1 1 1 1)))))
;(tree->meta '(? (((4 4) (2 1 1)))))
;(tree->meta '(? (((4 4) (2 2 1)))))
;(tree->meta '(? (((4 4) (1 (1 (1 1 1)))))))
;(tree->meta '(? (((4 4) (1 1 (2 (1 1 1)))))))

;(defun process-tree (obj &optional denominator)
;  "e.g. (process-tree '(1 (2 (1 (2 1)) 2)))"
;  (cond ((numberp obj) (leaf->entry obj denominator))
;        ((tree-p obj)
;         (list (or denominator (car obj))
;               (mapcar #'(lambda (x)
;                           (process-tree x ;leaf-value
;                                         (/ (round-duration (/ (/ (or denominator (car obj))
;                                                                  (note-list-sum (cadr obj))
;                                                                  ))))))
;                       (cadr obj))))))

(defun dur->notation (dur)
  "express any non-tuplet duration as multiples of a notatable unit
  e.g. (dur->notation 7/2)"
  (list (numerator (/ dur 16))
        (duration->value (/ dur (numerator (/ dur 16))))))

(defun notated-dur (tuplet)
  (rationalize (apply #'+ (mapcar #'(lambda (x)
                                      (rhythm (car x)))
                                  (cadr tuplet)))))

(defun tuplet-data (notated-dur occupied-dur)
  "e.g. (tuplet-data 3 2)"
  (if (> notated-dur occupied-dur)
      (append (dur->notation notated-dur)
              (dur->notation occupied-dur))))

(defun entryp (obj)
  "e.g. (entryp '(S NIL NIL NIL BREAK-BEAM NOTES))"
  (and (listp obj)
       (symbolp (car obj))))

(defun caadr-level (cons &optional (level 0))
  "e.g. (caadr-level (process-tree '(1 ((1 (1 1 1)) 1 1))))"
  (if (entryp cons)
      level
      (caadr-level (caadr cons)
                   (1+ level))))

(defun make-caadrs (n obj)
  "e.g. (make-caadrs 3 '(a b c))"
  (if (= n 0)
      obj
      (make-caadrs (1- n) (list 'caadr obj))))

(defun break-beam (tuplet)
  (setf
   (fifth (eval (make-caadrs (caadr-level tuplet)
                             `(quote ,tuplet))))
  'break-beam)
  tuplet)

(defun break-beams (x)
  (list (car x)
        (mapcar #'break-beam (cadr x))))

(defun give-tuplet (tuplet)
  (let ((notated-dur (notated-dur tuplet)))
    (unless (= (/ notated-dur 4) (car tuplet))
      (push
       (tuplet-data notated-dur (* 4 (car tuplet)))
       (fourth (eval (make-caadrs (caadr-level tuplet)
                                  `(quote ,tuplet)))))))
  tuplet)

;(defun give-tuplet (tuplet)
;  (let ((notated-dur (notated-dur tuplet)))
;    (unless (= (/ notated-dur 4) (car tuplet))
;      (let ((new-tuplet (tuplet-data notated-dur (* 4 (car tuplet)))))
;        (if new-tuplet 
;            (push (fourth (eval (make-caadrs (caadr-level tuplet)
;                                             `(quote ,tuplet))))
;                  tuplet)))))
;      tuplet)

; (give-tuplet '(1 ((Q NIL NIL NIL BREAK-BEAM NOTES) (E NIL NIL NIL BREAK-BEAM NOTES))))

(defun tupletp (obj)
  (and (listp obj)
       (numberp (car obj))
       (every #'(lambda (x) 
                  (or (entryp x)
                      (tupletp x)))
              (cadr obj))))

(defun remove-blank-tuplet (entry)
  "e.g. (remove-blank-tuplet '(Q NIL NIL (nil) BREAK-BEAM NOTES))"
  (if (equal (fourth entry) '(nil))
      (setf (fourth entry) nil))
  entry)

(defun tupletize (x)
  (cond ((entryp x) (remove-blank-tuplet x))
        ((tupletp x) (mapcar #'tupletize
                             (cadr (give-tuplet x))))))

;(defun tupletize (x)
;  (cond ((entryp x) x)
;        ((tupletp x) (mapcar #'tupletize 
;                             (cadr (give-tuplet x))))))

;(tupletize '(1 ((Q NIL NIL NIL BREAK-BEAM NOTES))))

; (tupletize (process-tree '(1 (2 (1 (2 1)) 2))))
; (dtrace::dtrace tupletize)

(defun flatten-entries (obj &optional accumulation)
  "e.g. (flatten-entries
 '(((Q NIL NIL NIL BREAK-BEAM NOTES) (Q NIL NIL NIL BREAK-BEAM NOTES))
            ((Q TIED-TO NIL NIL BREAK-BEAM NOTES) (Q NIL NIL NIL BREAK-BEAM NOTES))))"
  (cond ((null obj) accumulation)
        ((entryp obj) (cons obj accumulation))
        (t (flatten-entries (car obj)
                            (flatten-entries (cdr obj) accumulation)
                            ))))

(defun tie-meta (meta)
  "e.g. (tie-meta (tree->meta '(? (((4 4) ((1 (1 1 1)) 3.0)) ((4 4) (3.0 -1))))))"
  (let ((last-tie nil))
    (dolist (measure (reverse meta) meta)
      (dolist (entry (reverse measure))
        (if last-tie (setf (third entry) 'tied-from))
        (cond ((equal (second entry) 'tied-to) (setq last-tie T))
              (T (setq last-tie nil)))))))

(defun tree->meta (tree)
  "(tree->meta '(? (((4 4) ((1 (1 1 1)) 3.0)) ((4 4) (3.0 -1)))))"
  (tie-meta
   (mapcar #'(lambda (measure)
               (flatten-entries
                (tupletize
                 (break-beams
                  (process-tree (list (/ (caar measure)
                                         (cadar measure))
                                      (cadr measure)))))))
           (cadr tree))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;        SCORE->META         ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun mean-average (list-of-numbers)
  "e.g. (mean-average '(5 9 20))"
  (/ (apply #'+ list-of-numbers)
     (length list-of-numbers)))

(defun find-average-pitch (entry-list)
  "e.g. (find-average-pitch '(((h c4) (h cs4)) ((h nil) (h ef4))))"
  (mean-average (mapcar #'keynum 
                        (remove-if #'null 
                                   (mapcar #'cadr 
                                           (reduce #'append 
                                                   entry-list))))))

(defun find-middle-pitch (chords)
  "e.g. (find-middle-pitch '((60 61 70 90) (73 c3) nil))"
  (let ((numbers (reduce #'append chords)))
    (if (every #'null numbers)
      nil
      (let ((lowest (reduce #'min numbers)))
        (+ lowest (/ (- (reduce #'max numbers) lowest) 2))))))

(defun pick-clef (chords)
  "e.g. (pick-clef '((10 11 12) (73 c4) nil nil))"
  (let* ((chord-keynums (mapcar #'(lambda (x)
                                    (if x (keynum x)))
                                chords))
         (middle-pitch (find-middle-pitch chord-keynums)))
    (if (and middle-pitch
             (< (find-middle-pitch chord-keynums) 60))
        'bass
        'treble)))

(defun note-class (note-name)
  "e.g. (note-class 'bs)"
  (values (read-from-string (subseq (string note-name) 0 1))))

(defun last-instance-of-step-was-alteredp (note chord-list)
  "e.g. (last-instance-of-step-was-alteredp 'c5 '((d2) (b4 cs5)))"
    (find-if #'note-accidental 
             (remove-if-not #'(lambda (pitch-in-chord) 
                                (= (steps-from-middle-c note) (steps-from-middle-c pitch-in-chord)))
                            (find-if #'(lambda (chord)
                                         (member (steps-from-middle-c note)
                                                 (mapcar #'steps-from-middle-c chord)))
                                     chord-list))))

(defun chord-contains-altered-instance-of-note-namep (note chord)
  "e.g. (chord-contains-altered-instance-of-note-namep 'c4 '(c4 gs4 cs5))"
  (find-if #'(lambda (pitch-in-chord)
               (and (equal (note-class note) (note-class pitch-in-chord))
                    (note-accidental pitch-in-chord)))
           chord))

(defun last-instance-of-note-name-was-alteredp (note chord-list)
  "e.g. (last-instance-of-note-name-was-alteredp 'c5 '((d2) (b4 c5)))"
  (chord-contains-altered-instance-of-note-namep
   note
   (find-if #'(lambda (chord) (member (note-class note) (mapcar #'note-class  chord))) chord-list)))

;; Rules For Spelling:
;;
;; Show all sharps and flats unless the note is tied-to
;;
;; Show naturals only if 
;;  - same note-name appears altered in chord
;;  - OR last occurence of step in releveant region was non-natural
;;  - OR last chord in relevant region that contains that 
;;       note-name has any non-natural versions of that note-name
;;
;; not yet implemented:
;;  Hide accidentals if an event has exactly the same pitches as the
;;  previous event in the measure, and the previous event showed that
;;  accidental

(defun spell-meta (meta)
  (let ((current-measure nil)
        (previous-measure nil))
    (mapc 
     #'(lambda (measure)
         (mapc
          #'(lambda (event)
              (if (some #'numberp (sixth event))
                  (setf (sixth event) (spell (sixth event))))
              (setf (sixth event)
                    (mapcar
                     #'(lambda (pitch)
                         (list pitch
                               (cond ((second event) nil) ; if pitch is tied-to, don't show accidental
                                     ((note-accidental pitch) 'show-acci) ; show acci if pitch is non-natural
                                     ((chord-contains-altered-instance-of-note-namep pitch (sixth event)) 'show-acci)
                                     (T (let ((relevant-region (append current-measure previous-measure)))
                                          (cond ((last-instance-of-step-was-alteredp pitch relevant-region) 'show-acci)
                                                ((last-instance-of-note-name-was-alteredp pitch relevant-region) 
                                                 'show-acci)))))))
                     (sixth event)))
              (push (mapcar #'car (sixth event)) current-measure))
          measure)
         (setf previous-measure current-measure)
         (setf current-measure nil))
     meta)
    meta))

;this one doesn't fuse pauses:
(defun measures->simple2 (measures)
  (let* ((proportions (mapcar #'(lambda (measure)
                                  (om::unpack-tree (list (caar measure)
                                                     (cadr measure))))
                              measures))
         (multipliers (mapcar #'(lambda (measure proportion)
				  (/ (apply
				      '/ (car measure))
				     (truncate (apply '+ (mapcar 'abs proportion)))))
                              measures proportions))
         (global-proportion-list (om::flat (om::om* proportions (om::make-multiplier-list multipliers))))
         (global-multiplier (/ (first multipliers)
			       (first (om::make-multiplier-list multipliers)))))
    (om::om* global-proportion-list global-multiplier)))

(defun tree->onsets (tree)
  "e.g. (tree->onsets '(? (((4 4) (3 (1 (1 1 1)))) ((4 4) (-4)))))"
  (butlast (om::dx->x 0 (mapcar #'(lambda (tree-obj)
                                (* 4 (abs (rationalize tree-obj))))
                            (measures->simple2 (cadr tree))))))

(defun tree->time-sigs (tree)
  "e.g. (tree->time-sigs '(? (((4 4) (3 (1 (1 1 1)))) ((4 4) (-4)))))"
  (mapcar #'car (car (cdr tree))))

(defun give-register-maybe (chord)
  "e.g. (give-register-maybe '(70 0))"
  (mapcar #'(lambda (pitch-or-pc)
              (if (and (numberp pitch-or-pc)
                       (< pitch-or-pc 12))
                  (+ pitch-or-pc 60)
                  pitch-or-pc))
          chord))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun maybe-strip-nuances (time-instant event)
  (cond ((equal (car event) time-instant)
         (nthcdr 2 event))
        ((null event) '(nil))
        (t (list (third event)))))

(defun strip-rhythm (time-instant score)
  ;; e.g. (strip-rhythm 0 '((0 2 (60 62) ("2 4 1")) (2 4 (61 63)) (6 1 (70 71))))
  (maybe-strip-nuances time-instant
                       (find-if #'(lambda (event)
                                    (and (<= (first event) time-instant)
                                         (> (+ (first event) (second event)) time-instant)))
                                score)))

(defun process-meta (tree chords-and-nuances)
  "inserts pitch and nuances info into the tree"
  ;; e.g. (process-meta '(OM::? (((2 2) (1 1)) ((2 2) (1.0 #)))) '(((60 62) "help") ((61 63)) ((61 63)) ((70 71)) NIL)
  (mapcar #'(lambda (measure)
              (mapcar #'(lambda (event)
                          (append (butlast event)
                                  (pop chords-and-nuances)))
                      measure))
          (tree->meta tree)))

(defun make-meta (tree chords-and-nuances)
  "adds time signatures and clefs"
  (list (tree->time-sigs tree)
        (append (list (pick-clef (mapcar #'car chords-and-nuances)))
                (spell-meta (process-meta tree chords-and-nuances)))))

(defun score->meta (score &optional (time-sigs '((4 4))))
  ;; (score->meta '((0 2 (60 62) ("help" "crap")) (2 4 (61 63)) (6 1 (70 71))))
  (let* ((tree (score->tree score time-sigs))
         (chords-and-nuances (mapc #'(lambda (x)
                                       (unless (null x)
                                         (setf (first x)
                                               (give-register-maybe (first x)))))
                                   (mapcar #'(lambda (onset)
                                               (strip-rhythm onset score))
                                           (tree->onsets tree)))))
    (make-meta tree chords-and-nuances)))

;; (dtrace:duntrace)
;; (dtrace:dtrace strip-rhythm)
;; (dtrace:duntrace spell-meta)
;; (dtrace:dtrace process-meta)
;; (dtrace:duntrace make-meta)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; the original functions:

; (defun make-meta (tree chords)
;   (list (tree->time-sigs tree)
;         (append (list (pick-clef chords))
;                 (spell-meta
;                  (mapcar #'(lambda (measure)
;                              (mapcar #'(lambda (event)
;                                          (append (butlast event)
;                                                  (list (pop chords))))
;                                      measure))
;                          (tree->meta tree))))))

; (defun score->meta (score &optional (time-sigs '((4 4))))
;   "e.g. (score->meta '((0 2 (60 62) ) (2 4 (61 63)) (6 1 (70 71))) '(2 2))"
;   (let* ((tree (score->tree score time-sigs))
;          (chords (mapcar #'(lambda (onset)
;                              (give-register-maybe (sounding-pitches onset score)))
;                          (tree->onsets tree))))
;     (make-meta tree chords)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (defun make-meta (tree chords-and-nuances)
;   (list (tree->time-sigs tree)
;         (append (list (pick-clef (mapcar #'car chords-and-nuances)))
;                 (spell-meta
;                  (mapcar #'(lambda (measure)
;                              (mapcar #'(lambda (event)
;                                          (append (butlast event)
;                                                  (pop chords-and-nuances)))
;                                      measure))
;                          (tree->meta tree))))))

; (defun score->meta (score &optional (time-sigs '((4 4))))
;   ;; (score->meta '((0 2 (60 62) "help") (2 4 (61 63)) (6 1 (70 71))) '(2 2))
;   (let* ((tree (score->tree score time-sigs))
;          (chords-and-nuances (mapc #'(lambda (x)
;                                        (unless (null x)
;                                          (setf (first x)
;                                                (give-register-maybe (first x)))))
;                                    (mapcar #'(lambda (onset)
;                                                (strip-rhythm onset score))
;                                            (tree->onsets tree)))))
;     (make-meta tree chords-and-nuances)))




; (make-meta '(? (((4 4) (1 (2 (1 -1 1)) 1.0)))) '((60) (cs4 ef4) nil (60 61) (60 61)))

; (pick-clef '((60) (cs4 ef4) nil (60 61) (60 61)))

; (open-finale)

; (spell '(cs4 ef4))

;(defun score->meta (score &optional (time-sigs '((4 4))))
;  "e.g. (score->meta '((0 2 (60 62)) (2 4 (61 63)) (6 1 (70 71))) '(2 2))"
;  (let* ((tree (score->tree score time-sigs))
;         (time-sigs (tree->time-sigs tree))
;         (chords (mapcar #'(lambda (onset) 
;                             (give-register-maybe (sounding-pitches onset score)))
;                         (tree->onsets tree)))
;         (clef (pick-clef chords)))
;    (list time-sigs
;          (append (list clef)
;                 (spell-meta
;                   (mapcar #'(lambda (measure)
;                               (mapcar #'(lambda (event)
;                                           (append (butlast event) (list (pop chords))))
;                                       measure))
;                           (tree->meta tree)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;         MAKE-META          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (tree->meta '(? (((4 4) ((1 (1 1 1)) 1 1)))))
; (make-meta '(? (((4 4) ((1 (1 1 1)) 1 1)))) '((60) (60) (60) (60) (60)))

; (tree->meta '(? (((4 4) (1 1 (2 (1 1 1)))))))

;(cm-etf::meta->etf
; '(((4 4))
;   (TREBLE
;    ((Q NIL NIL ((3 Q 1 H) (3 H 1 W)) NIL ((CM::C4 NIL)))
;     (Q NIL NIL NIL NIL ((CM::C4 NIL)))
;     (Q NIL NIL NIL NIL ((CM::C4 NIL)))
;     (H NIL NIL NIL NIL ((CM::C4 NIL)))
;     (H NIL NIL NIL NIL ((CM::C4 NIL)))))))

;(make-meta '(? (((4 4) (1 1 (2 (1 1))))))
;           '((60) (60) (60) (60) (60)))


;(om::mktree '(1/4 1/4 1/6 1/6 1/6) '((4 4)))

;(open-finale)

;(in-package :cm-etf)

;(events (new seq :subobjects (list (new midi :time 0 :duration 4/9)
;                                   (new midi :time 4/9 :duration 4/9)
;                                   (new midi :time 8/9 :duration 4/9)
;                                   (new midi :time 12/9 :duration 4/3)
;                                   (new midi :time 8/3 :duration 4/3)))
;        "c:/lisp/test-tuplet.ly")

; fix round-duration bug in 
; rewrite spell-meta separating spelling and cautionaries
; export symbols in packages
; rewrite README
; add docstring to make-etf

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;         META->ETF          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun meta->score-setup (meta &optional (measures-per-system 3))
  "e.g. (meta->score-setup *test-meta*)"
  (let* ((time-sig-list (car meta))
         (number-of-measures (length time-sig-list))
         (clef-list (mapcar #'car (cdr meta)))
         (number-of-systems (ceiling number-of-measures measures-per-system))
         (systems-per-page (systems-per-page (length clef-list)))
         (system-width 2160))
  (enter-tags (make-staves clef-list))
  (enter-tags (make-staff-positions clef-list))
  (enter-tags (make-systems number-of-measures measures-per-system))
  (enter-tags (make-pages number-of-systems systems-per-page))
  (enter-tags (make-measures2003 time-sig-list measures-per-system system-width))))


;; (score->meta '((0 2 (60 62) ("help" "crap")) (2 4 (61 63)) (6 1 (70 71))))

; (((4 4) (4 4))
;  (TREBLE
;   ((H NIL NIL NIL BREAK-BEAM ((CM::C4 NIL) (CM::D4 NIL)) ("help" "crap"))
;    (H NIL TIED-FROM NIL BREAK-BEAM ((CM::DF4 SHOW-ACCI) (CM::EF4 SHOW-ACCI))))
;   ((H TIED-TO NIL NIL BREAK-BEAM ((CM::DF4 NIL) (CM::EF4 NIL)))
;    (Q NIL NIL NIL BREAK-BEAM ((CM::AS4 SHOW-ACCI) (CM::B4 NIL))) (Q NIL NIL NIL BREAK-BEAM NIL))))



; ((0 2 (60 62) ("help" "crap") (2 4 (61 63)) (6 1 (70 71))) '(2 2))

; (defun extract-expressions (meta)
;   (

;; (apply #'score->meta (car (mapcar #'cadr (cm::indexed-grid '(2 1) 4))))


; (defun make-expression-definitions (


(defun steps-from-middle-c (cmn-note)
  "e.g. (steps-from-middle-c 'b3)"
  (second (decode-interval (interval 'c4 cmn-note))))

(defun consecutive-integers (number-list &optional (result (list (car number-list))))
  "e.g. (consecutive-integers '(8 6 5 4 2))"
  (let ((2nd (cadr number-list)))
    (cond ((or (null 2nd) 
               (> (abs (- 2nd (car result))) 1)) 
           (reverse result))
          (T (consecutive-integers (cdr number-list) (push 2nd result))))))

(defun upstem-displacep (n interval-list)
  "e.g. (upstem-displacep 5 '(-10 -5 -4 3 7 8 9 11))"
  (let ((the-cluster (consecutive-integers (reverse (subseq interval-list 0 (1+ n))))))
    (if (oddp (- (car (last the-cluster)) (car the-cluster))) 'upstem-displace)))

(defun downstem-displacep (n interval-list)
  "e.g. (downstem-displacep 1 '(-10 -5 -4 3 7 8 9 11))"
  (let ((the-cluster (consecutive-integers (subseq interval-list n))))
    (if (oddp (- (car (last the-cluster)) (car the-cluster))) 'downstem-displace)))

(defun meta-entry->entry (entry-number meta-entry)
  "e.g. (meta-entry->entry 1 '(E NIL NIL (3 E 2 E) break-beam ((C4 NIL) (D4 NIL) (E4 NIL) (FS4 SHOW-ACCI)) NIL))"
  (let* ((result nil)
         (pitch-list (sort (sixth meta-entry) #'(lambda (x y) (< (keynum x) (keynum y))) :key #'car))
         (interval-list (mapcar #'(lambda (x) (steps-from-middle-c (car x))) pitch-list)))
    (push (entry-spec entry-number
                      (1- entry-number) ;prev-entry
                      (1+ entry-number) ;next-entry
                      (translate-rhythm (car meta-entry)) ;duration in EDU's
                      (entry-flag (if pitch-list 'note)       ;note or rest
                                  (listp (fourth meta-entry)) ;entry begins a tuplet
                                  (if (fifth meta-entry) 'break-beam))
                      (if (seventh meta-entry) 'note-expr) ;non-nil indicates the presence of note expression(s) for that entry
                      (length pitch-list))
          result)
    (dotimes (i (length pitch-list))
      (push (note-record (car (nth i pitch-list)) ;cmn-note
                         (1+ i)         ;note-record number
                         (if (equal (second (nth i pitch-list)) 'show-acci) T) ;show-accidental
                         (if (equal (second (nth i pitch-list)) 'show-acci) T) ;freeze accidental
                         nil            ;don't parenthesize accidental
                         (if (equal (second meta-entry) 'tied-to) T) ;tied-to
                         (if (equal (third meta-entry) 'tied-from) T) ;tied-from
                         nil            ;don't cross-staff
                         (downstem-displacep i interval-list) ;displace notehead for downstem
                         (upstem-displacep i interval-list)) ;displace notehead for upstem
            result))
    (reduce #'(lambda (x y) (concatenate 'string x y)) (reverse result))))

(defun meta-entry->tuplet (entry-number meta-tuplet)
  "e.g. (meta-entry->tuplet 1 '(3 e 2 e))
  (meta-entry->tuplet 1 '(3 h 1 b))"
  (tuplet-spec entry-number
               (first meta-tuplet)
               (translate-rhythm (second meta-tuplet))
               (third meta-tuplet)
               (translate-rhythm (fourth meta-tuplet))))

(defun meta->entry-info (meta)
  "e.g. (meta->entry-info *test-meta*)"
  (let ((entries (list (format nil "~%^entries ~%")))
        (tuplets nil)
        (frames nil)
        (frame-refs nil)
        (staff-number 0)
        (measure-number 0)
        (frame-number 0)
        (entry-number 1))
    (dolist (staff (cdr meta))
      (incf staff-number)
      (setf measure-number 0)
      (dolist (measure (cdr staff))
        (incf measure-number)
        (incf frame-number)
        (push (frame-spec frame-number entry-number (+ entry-number (1- (length measure)))) frames)
        (push (frame-ref-spec staff-number measure-number (car staff) frame-number) frame-refs)
        (dolist (entry measure)
          (push (meta-entry->entry entry-number entry) entries)
          (if (fourth entry)
              (dolist (x (fourth entry))
                (push (meta-entry->tuplet entry-number x) tuplets)))
          (incf entry-number))))
    (mapc #'(lambda (tag-list)
              (if tag-list (enter-tags (reduce #'(lambda (y z) 
                                                   (concatenate 'string y z))
                                               (reverse tag-list)))))
          `(,entries ,frames ,frame-refs ,tuplets))
    T))


(defun meta->entry-info (meta new-expression-list)
  "e.g. (meta->entry-info *test-meta*)"
  (let ((entries (list (format nil "~%^entries ~%")))
        tuplets
        frames
        frame-refs
        expression-details
        (expression-definition-offset (max-expression-number *etf-string*))
        (staff-number 0)
        (measure-number 0)
        (frame-number 0)
        (entry-number 1))
    (dolist (staff (cdr meta))
      (incf staff-number)
      (setf measure-number 0)
      (dolist (measure (cdr staff))
        (incf measure-number)
        (incf frame-number)
        (push (frame-spec frame-number entry-number (+ entry-number (1- (length measure)))) frames)
        (push (frame-ref-spec staff-number measure-number (car staff) frame-number) frame-refs)
        (dolist (entry measure)
          (push (meta-entry->entry entry-number entry) entries)
          (if (fourth entry)
              (dolist (x (fourth entry))
                (push (meta-entry->tuplet entry-number x) tuplets)))
          (if (seventh entry)
              (dolist (expression (seventh entry))
                (push (make-expression-details entry-number
                                               (+ (expression-number expression new-expression-list)
                                                  expression-definition-offset))
                      expression-details)))
          (incf entry-number))))
    (mapc #'(lambda (tag-list)
              (if tag-list (enter-tags (reduce #'(lambda (y z) 
                                                   (concatenate 'string y z))
                                               (reverse tag-list)))))
          `(,entries ,frames ,frame-refs ,tuplets ,expression-details))
    T))


; this would only be needed for beat-charts:
; (defun tuplet->ratio (tuplet)
;   "e.g. (tuplet->ratio '(3 e 2 e))"
;   (/ (* (third tuplet)
;         (rationalize (rhythm (fourth tuplet) 60)))
;      (* (first tuplet)
;         (rationalize (rhythm (second tuplet) 60)))))

; (defun meta->etf (meta &optional (filename *enigma-output-filename*) &key new-expressions)
;   "e.g. (meta->etf *test-meta*)"
;   (reset-etf)
;   (meta->score-setup meta)
;   (if new-expressions (enter-tags (expand-expression-definitions new-expressions)))
;   (meta->entry-info meta)
;   (save-etf filename))


(defun meta->etf (meta &key new-expressions (filename *enigma-output-filename*))
  "e.g. (meta->etf *test-meta*)"
  (reset-etf)
  (meta->score-setup meta)
  (if new-expressions (enter-tags (expand-expression-definitions new-expressions)))
  (meta->entry-info meta new-expressions)
  (save-etf filename))

; (score->etf '((0 1 (60 63) ("help" "crap")) (1 4 (61 65) ("crappy-crap")) (5 3 (62 66)))
;             '((0 3 (48 51)) (3 4 (49 53) ("hell")) (7 1 (50 54))))

; (open-finale)

; (dtrace::dtrace meta->etf)


; '(((4 4) (4 4))
;  (TREBLE
;   ((Q NIL NIL NIL BREAK-BEAM ((CM::C4 NIL) (CM::EF4 SHOW-ACCI)) ("help" "crap"))
;    (H. NIL TIED-FROM NIL BREAK-BEAM ((CM::DF4 SHOW-ACCI) (CM::F4 NIL)) ("crappy-crap")))
;   ((Q TIED-TO NIL NIL BREAK-BEAM ((CM::DF4 NIL) (CM::F4 NIL)))
;    (H. NIL NIL NIL BREAK-BEAM ((CM::D4 SHOW-ACCI) (CM::FS4 SHOW-ACCI)))))
;  (BASS
;   ((H. NIL NIL NIL BREAK-BEAM ((CM::C3 NIL) (CM::EF3 SHOW-ACCI)))
;    (Q NIL TIED-FROM NIL BREAK-BEAM ((CM::DF3 SHOW-ACCI) (CM::F3 NIL)) ("hell")))
;   ((H. TIED-TO NIL NIL BREAK-BEAM ((CM::DF3 NIL) (CM::F3 NIL)))
;    (Q NIL NIL NIL BREAK-BEAM ((CM::D3 SHOW-ACCI) (CM::FS3 SHOW-ACCI))))))



; (open-finale)

(defun make-etf (tree chords)
  "MAKE-ETF writes an ETF based on a rhythm tree 
 and a list of chords. This allows the user to bypass 
 the CM-ETF quantization and spelling algorithms.

 e.g. (make-etf '(? (((4 4) (1 (2 (1 -1 1)) 1.0))))
                '((60) (cs4 ef4) nil (60 61) (60 61)))"
  (meta->etf (make-meta tree chords)))

; (defvar *test-meta* 
;   '(((4 4) (4 4))
;     (TREBLE
;      ((E NIL NIL ((3 E 1 Q)) BREAK-BEAM ((C4 NIL) (FS4 SHOW-ACCI)))
;       (E NIL NIL NIL NIL NIL)
;       (E NIL TIED-FROM NIL NIL ((F4 SHOW-ACCI)))
;       (H. TIED-TO TIED-FROM NIL BREAK-BEAM ((F4 NIL))))
;      ((H. TIED-TO NIL NIL BREAK-BEAM ((F4 NIL)))
;       (Q NIL NIL NIL BREAK-BEAM NIL)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;        SCORE->ETF          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun extract-sigs (list)
  "e.g. (extract-sigs '(a b :time-sigs d e))"
  (if (find ':time-sigs list)
      (nth (1+ (search (list ':time-sigs) list)) list)))

(defun remove-sigs (list)
  "e.g. (remove-sigs '(a b :time-sigs d e))"
  (remove ':time-sigs (remove (extract-sigs list) list)))

(defun get-expressions (big-list)
;; e.g. (get-expressions '(((0 1 (60 63) ("help" "crap")) (1 4 (61 65) ("crappy-crap")) (5 3 (62 66))) ((0 3 (48 51)) (3 4 (49 53) ("hell" "crap")) (7 1 (50 54)))))"
  (reduce #'(lambda (x y)
              (union x y :test #'string-equal))
   (mapcar #'(lambda (event)
               (reduce #'(lambda (x y)
                           (union x y :test #'string-equal))
                       (mapcar #'fourth event)))
           big-list)))

;; (get-expressions cm::*final-grid*)
;; (in-package :cm)

(defun score->etf (&rest scores-and-maybe-time-sigs)
  "SCORE->ETF writes lists of events to ETF. 
The format of each event is: (time duration (pitches)).  
Each event list supplied as an argument is written to a separate staff.  
Can specify time signatures with :time-sigs keyword.
e.g. (score->etf '((0 1 (60 63)) (1 4 (61 65)) (5 3 (62 66)))
                 '((0 3 (48 51)) (3 4 (49 53)) (7 1 (50 54)))
                 :time-sigs '((2 2) (3 4)))"
  (let ((expression-defs-to-add (get-expressions scores-and-maybe-time-sigs))
        (big-list (mapcar #'(lambda (score)
                              (score->meta score (or (extract-sigs scores-and-maybe-time-sigs) '(4 4))))
                          (remove-sigs scores-and-maybe-time-sigs))))
;    expression-defs-to-add))
    (meta->etf (append (list (car (sort (mapcar #'car big-list) #'> :key #'length)))
                       (mapcar #'second big-list))
               :new-expressions expression-defs-to-add)))

; (score->etf '((0 1 (60 63) ("help" "crap")) (1 4 (61 65) ("crappy-crap")) (5 3 (62 66)))
;             '((0 3 (48 51)) (3 4 (49 53) ("hell" "help")) (7 1 (50 54))))

; (open-finale)

; (dtrace::test-thingy)
; (dtrace::dtrace meta->etf)

; (expand-expression-definitions '("help" "crap" "crappy-crap" "hell"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;         SEQ->ETF           ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; first, consolidate multiple SEQs into a single SEQ:

(defun separate-seq-channels (seq)
  "e.g. (separate-seq-channels
         (new seq :subobjects
              (list
               (new midi :time 0 :duration 8/3 :keynum 61 :channel 0)
               (new midi :time 8/3 :duration 4/3 :keynum 62 :channel 0)
               (new midi :time 0 :duration 8/3 :keynum 61 :channel 1)
               (new midi :time 8/3 :duration 4/3 :keynum 62 :channel 8))))"
  (let ((seq-subobjects (remove-if-not #'(lambda (x) 
                                           (eql 'midi (type-of x)))
                                       (subobjects seq))))
    (mapcar #'(lambda (channel)
                (remove-if-not #'(lambda (event)
                                   (= channel (slot-value event 'channel)))
                               seq-subobjects))
            (used-channels seq))))

(defun next-channel (channel-number)
  "e.g. (next-channel 8)"
  (if (= (mod (1+ channel-number) 16) 9) (+ 2 channel-number) (1+ channel-number)))

(defun superimpose-seqs (&rest seqs)
  (let ((result-seq (new seq))
        (channel 0)
        (big-list (apply #'append (mapcar #'separate-seq-channels seqs))))
    (mapc #'(lambda (channel-group)
              (mapc #'(lambda (event)
                        (insert-object (new midi
                                            :time (slot-value event 'time)
                                            :duration (slot-value event 'duration)
                                            :keynum (slot-value event 'keynum)
                                            :amplitude (slot-value event 'amplitude)
                                            :channel channel)
                                       result-seq))
                    channel-group)
              (setf channel (next-channel channel)))
          big-list)
    result-seq))

(defun seq->etf (&rest seqs-and-maybe-time-sigs)
  "SEQ->ETF translates a SEQ object and writes it to
 `cm-etf_OUTPUT.ETF' in the CM-ETF directory.
 Each channel is written to a separate staff.
 e.g. (seq->etf (new seq :subobjects 
                     (list (new midi :time 0 :duration 3 :keynum 61)
                           (new midi :time 3 :duration 1 :keynum 62)
                           (new midi :time 3 :duration 1 :keynum 63)
                           (new midi :time 0 :duration 4 :keynum 64 :channel 1))))"
  (apply #'score->etf 
         (append (seq->scores (apply #'superimpose-seqs 
                                     (remove-sigs seqs-and-maybe-time-sigs)))
                 (list :time-sigs 
                       (or (extract-sigs seqs-and-maybe-time-sigs) 
                           '(4 4))))))
