;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-

;;; CM-ETF: music notation output in Finale format from Common Music
;;; Copyright (C) 2006 Niklas Kambeitz (niklas.kambeitz@mail.mcgill.ca)
;;;
;;; last modified Time-stamp: <07/05/18 20:57:26 nik>
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :cm-etf)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;  STAVES, SYSTEMS & PAGES   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun translate-clef (clef-name)
  "e.g. (translate-clef 'bass)"
  (cadr (assoc clef-name '((treble 0) (bass 3)))))

;(get-tag "^IS" *etf-string*)
;
; Between Finale 2000 and 2003, lines 4, 5 and 6 were added to this.
;
(defun staff-spec (staff-number clef)
  "e.g. (staff-spec 1 'bass)"
  (format nil "~%^IS(~s) ~s ~s ~s ~s ~s ~s 
^IS(~s) ~s ~s ~s ~s ~s ~s 
^IS(~s) ~s ~s ~s ~s ~s ~s 
^IS(~s) ~s ~s ~s ~s 
^IS(~s) ~s ~s ~s 
^IS(~s) ~s ~s ~s " 
          
          staff-number ;this whole line is uncertain due to documentation errors
          0     ;botBarlineOffset: 0 = measure width is 600 EVPUs)
          0     ;baseyoff - only used for tab staves
          26368 ;mfont - for independent notehead font 
          0     ;sizeefx - for independent notehead font
          0     ;flag:  0 = nothing special
          0     ;unknown

          staff-number
          (translate-clef clef) ;0 = treble, 3 = bass
          0    ;topLines: 0 = normal, non-custom staff
          5    ;botLines: 5 = five staff lines
          0    ;topBarlineOffset
          0    ;transposition: 0 = no transposition
          0    ;instflag

          staff-number
          -772 ;dw_wRest: -772 (0xfcfc) = all rests are placed -4 steps from top line
          -772 ;h_otherRest: -772:  rests are placed -4 steps from top line
          -4   ;stemReversal: -4 = stem reversal line is -4 steps from top line
          0    ;fullName: 0 = no text blocks for full staff name
          0    ;abbrvName: 0 = no text blocks for abbreviated staff name
          -517 ;unknown

          staff-number
          1536  ;unknown
          -1024 ;unknown
          9247  ;unknown 
          0     ;unknown

          staff-number
          0    ;unknown - probably unused
          0    ;unknown - probably unused
          0    ;unknown - probably unused

          staff-number
          0    ;unknown - probably unused
          0    ;unknown - probably unused
          0))  ;unknown - probably unused

(defun make-staves (staff-list)
  "e.g. (make-staves '(treble bass))"
  (let (result)
    (dotimes (i (length staff-list) result)
      (setf result (concatenate 'string result (staff-spec (1+ i) (nth i staff-list)))))))

;(get-tag "^Iu" *etf-string*)
;
; This tag is not documented.
;
(defun staff-position (staff-number staff-position)
  "e.g. (staff-position 1 0)"
  (format nil "~%^Iu(0) ~s 0 0 0 ~s ~%^Iu(0) 1 0 32767 2147483647 " 
          staff-number
          staff-position))

(defun make-staff-positions (staff-list)
  "e.g. (make-staff-positions '(treble bass))"
  (let (result)
    (dotimes (i (length staff-list) result)
      (setf result (concatenate 'string result (staff-position (1+ i) (* i -320)))))))

;(get-tag "^MS" *etf-string*)
;
; ETFSPEC says there's only one instance for each measure; Finale 2005 makes three.
;
(defun measure-spec (measure-number measure-width key-sig number-of-beats beat-duration spacing-type barline-flag)
  "e.g. (measure-spec 1 600 0 4 1024 4 16)"
  (format nil "~%^MS(~s) ~s ~s ~s ~s ~s ~s ~%^MS(~s) ~s ~s ~s ~s ~s ~s ~%^MS(~s) ~s ~s ~s ~s ~s ~s "
          measure-number
          measure-width   ;in EVPU's
          key-sig         ;0 = C Major
          number-of-beats
          beat-duration   ;in EDU's: 1024 = quarter note
          spacing-type    ;4 = time-signature, 6 = beat-chart, 8198 = ? [auxflag]
          barline-flag    ;16 = normal barline type, 16400 = ?          [meflag]
 
          measure-number 
          number-of-beats ;possibly for "use a different time signature for display
          beat-duration   ;possibly for "use a different time signature for display
          240             ;unknown
          0               ;unknown
          0               ;unknown
          0               ;unknown

          measure-number
          0               ;unknown
          0               ;unknown
          0               ;unknown
          0               ;unknown
          0               ;unknown
          0))             ;unknown

(defun make-measures (number-of-measures measures-per-system system-width)
  "e.g. (make-measures 20 3 2160)"
  (let (result)
    (dotimes (i number-of-measures)
      (push (measure-spec (1+ i) ;measure number
                          (/ system-width measures-per-system) ;measure width --- stretched for last system?
                          0      ;key signature
                          4      ;number of beats
                          1024   ;beat duration
                          4      ;time signature spacing
                          16)    ;normal barline
            result))
    (reduce #'(lambda (x y) (concatenate 'string x y)) (reverse result))))

(defun measure-spec2003 (measure-number measure-width key-sig number-of-beats beat-duration spacing-type barline-flag)
  "e.g. (measure-spec2003 1 600 0 4 1024 4 16)"
  (format nil "~%^MS(~s) ~s ~s ~s ~s ~s ~s ~%^MS(~s) ~s ~s ~s ~s ~s ~s "
          measure-number
          measure-width   ;in EVPU's
          key-sig         ;0 = C Major
          number-of-beats
          beat-duration   ;in EDU's: 1024 = quarter note
          spacing-type    ;4 = time-signature, 6 = beat-chart, 8198 = ? [auxflag]
          barline-flag    ;16 = normal barline type, 16400 = ?          [meflag]

          measure-number 
          number-of-beats ;possibly for "use a different time signature for display
          beat-duration   ;possibly for "use a different time signature for display
          240             ;unknown
          0               ;unknown
          0               ;unknown
          0))             ;unknown

;; adding a note-attached text-expression changes the 'barline-flag' from 16 to 271 ;i.e. it adds 255

(defun make-measures2003 (time-sig-list measures-per-system system-width)
  "e.g. (make-measures2003 '((4 4) (4 4)) 2 2160)"
  (let (result)
    (dotimes (i (length time-sig-list))
      (push (measure-spec2003 (1+ i) ;measure-number
                              (/ system-width measures-per-system) ;measure-width --- stretched for last system?
                              0 ;key-signature
                              (car (nth i time-sig-list)) ;number of beats
                              (* 4096 (/ 1 (cadr (nth i time-sig-list)))) ;duration of beats
                              4 ;
                              16)
            result))
    (reduce #'(lambda (x y) (concatenate 'string x y)) (reverse result))))

;(get-tag "^SS" *etf-string*)
;
;Finale 2005 uses more data fields for staff specs 
;than are described in the etfspec for Finale 2000
;
(defun staff-system-spec (sys-number top-marg left-marg right-marg bottom-marg first-measure stave-flag last-measure horiz-percent resize-percent)
  "e.g. (staff-system-spec 1 -80 0 0 -200 1 3 4 306.50 100)"
  (format nil "~%^SS(~s) ~s ~s ~s ~s ~s ~s ~%^SS(~s) ~s ~s ~s ~s ~s ~s ~%^SS(~s) ~s ~s ~s ~s ~s ~s " 
          sys-number
          top-marg
          left-marg
          right-marg
          bottom-marg
          first-measure
          stave-flag ;1=hold_margins, 3=standard

          sys-number
          last-measure
          0                              ;horzPercentHi [only enormously long systems need this]
          (round (* horiz-percent 100))  ;horzPercentLo [a tyobyte = 65536 = 655.36%]
          resize-percent 
          0    ;unknown
          1536 ;unknown - used

          sys-number
          0    ;unknown - apparently useless
          0    ;unknown - probably unused
          0    ;unknown - probably unused
          0    ;unknown - probably unused
          0    ;unknown - probably unused
          0))  ;unknown - probably unused

(defun make-systems (number-of-measures measures-per-system)
  "e.g. (make-systems 20 3)"
  (let (result)
    (dotimes (i (ceiling number-of-measures measures-per-system))
      (push (staff-system-spec (1+ i) ;system number
                               -80    ;top margin
                               0      ;left margin
                               0      ;right margin
                               -200   ;bottom margin
                               (1+ (* i measures-per-system)) ;first measure in system
                               3      ;stave flag
                               (min (+ (1+ (* i measures-per-system)) measures-per-system) number-of-measures) ;last-meas
                               306.5  ;horizontal stretch percent ***NOTE***: this should be changing, not constant
                               100)   ;resize-percent
            result))
    (reduce #'(lambda (x y) (concatenate 'string x y)) (reverse result))))

;(get-tag "^PS" *etf-string*)

(defun page-spec (page-number height width first-system page-flag top-marg left-marg bottom-marg right-marg resize-percent)
  "e.g. (page-spec 1 3168 2448 1 2 -144 144 144 -144 80)"
  (format nil "~%^PS(~s) ~s ~s ~s ~s ~%^PS(~s) ~s ~s ~s ~s ~s 0 "
          page-number
          height
          width
          first-system
          page-flag ;2 means page was resized with "hold margins", no ossias
 
          page-number
          top-marg
          left-marg
          bottom-marg
          right-marg
          resize-percent))

(defun make-pages (number-of-systems systems-per-page)
  "e.g. (make-pages 20 3)"
  (let (result)
    (dotimes (i (ceiling number-of-systems systems-per-page))
      (push (page-spec (1+ i) 
                       3168 ;page height 
                       2448 ;page width
                       (1+ (* i systems-per-page))
                       2    ;hold margins
                       -144 ;top margin
                       144  ;left margin
                       144  ;bottom margin
                       -144 ;right margin
                       100) ;resize percent
            result))
    (reduce #'(lambda (x y) (concatenate 'string x y)) (reverse result))))

(defun systems-per-page (number-of-staves)
  "e.g. (systems-per-page 2)"
  (let ((page-space (- 3168             ;page height
                       144              ;top page margin
                       144))            ;bottom page margin
        (system-height (+ 96            ;staff height
                          80            ;top system margin
                          200           ;bottom system margin
                          (* number-of-staves 320)))) ;staff spacing
    (car (multiple-value-list (ceiling page-space system-height)))))

;got two bars per system instead of three!?

;NOT NECESSARY
;(get-tag "^PL" *etf-string*)
; (defun playback-spec (staff-number &optional GM-number)
;   "e.g. (playback-spec 1)"
;   (format nil "~%^PL(~s) 0 0 0 0 0 0 ~%^PL(1) 20565 0 0 0 0 0 " 
;           staff-number
;           staff-number))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;          ENTRIES           ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun translate-rhythm (cmn-rhythm)
  "e.g. (translate-rhythm 'q3..)"
  (values (round (* (rhythm cmn-rhythm 60) 1024))))

(defun duration->edu (dur)
  "e.g. (duration->edu 1)"
  (values (round (* dur 1024))))

(defun time-denom->edu (dur)
  "e.g. (time-denom->edu 4)"
  (* (/ 1 dur) 4096))

(defun value->edu (value)
  "e.g. (value->edu 'e)"
  (let ((value-alist '((M 32) (L 16) (B 8) (W 4) (H 2) (Q 1) (E 1/2) (S 1/4) (T 1/8) (X 1/16))))
    (cond ((listp value) 
           (* 1024 
              (cadr (assoc (car value) value-alist))
              (cadr (assoc (cadr value) '((0 1) (1 3/2) (2 7/4) (3 15/8) (4 31/16))))))
           (T (* 1024 (cadr (assoc value value-alist)))))))

(defun duration->value (dur)
  "e.g. (duration->value 31/32)"
  (or
   (cadr (assoc dur '((32 M) (16 L) (8 B) (4 W) (2 H) ; undotted durations
                      (1 Q) (1/2 E) (1/4 S) (1/8 T) (1/16 X) ;
                      (48 M.) (24 L.) (12 B.) (6 W.) (3 H.) ; durations with one dot
                      (3/2 Q.) (3/4 E.) (3/8 S.) (3/16 T.) (3/32 X.) ;
                      (56 M..) (28 L..) (14 B..) (7 W..) (7/2 H..) ; durations with 2 dots
                      (7/4 Q..) (7/8 E..) (7/16 S..) (7/32 T..) (7/64 X..) ;
                      (60 M...) (30 L...) (15 B...) (15/2 W...) (15/4 H...) ; durations with 3 dots
                      (15/8 Q...) (15/16 E...) (15/32 S...) (15/64 T...) (15/128 X...) ;
                      (62 M....) (31 L....) (31/2 B....) (31/4 W....) (31/8 H....) ; durations with 4 dots
                      (31/16 Q....) (31/32 E....) (31/64 S....) (31/128 T....) (31/256 X....)))) ;
   (error "DURATION->VALUE: No note value for ~s" dur)))

(defun translate-note-name (cmn-note-name)
  "e.g. (translate-note-name 'bf)"
  (* 16 (cadr (assoc (read-from-string (subseq (string (note-name cmn-note-name)) 0 1) 'symbol)
                     '((c 0) (d 1) (e 2) (f 3) (g 4) (a 5) (b 6))))))

(defun translate-alteration (cmn-accidental)
  "e.g. (translate-alteration 'ff)"
  (cond ((null cmn-accidental) 0)
        ((equal cmn-accidental 'n) 0)
        (t (+ (length (string cmn-accidental))
              (if (search "F" (string cmn-accidental)) 8 0)))))

(defun translate-octave (cmn-octave)
  "e.g. (translate-octave 3)"
  (* 112 (- cmn-octave 4)))

(defun translate-note (cmn-note)
  "e.g. (translate-note 'bn3)"
  (+ (translate-note-name (note-name cmn-note))
     (translate-alteration (note-accidental cmn-note))
     (translate-octave (octave-number cmn-note))))

(defun note-record (cmn-note note-id show-acci freeze-acci paren-acci tied-to tied-from cross-staff ds-2nd us-2nd)
  "e.g. (note-record 'bf3 9 'show-accidental 'freeze-accidental 'paren 'tied-to 'tied-from  'cross-staff 'ds-2nd 'us-2nd)"
  (format nil "~%    ~s $~x~x~x~x~x~x~x~x " 
          (translate-note cmn-note) ;"tone center displacement" (i.e. how many steps above or below middle C)
          (+ 8 ;what's this for?    ; everything here and below is called the "note-flag"
             (if tied-to 2 0) 
             (if tied-from 4 0)
             (if cross-staff 1 0))
          (+ (if show-acci 1 0) 
             (if ds-2nd 4 0) 
             (if us-2nd 8 0))  ;2 = goes on upper split stem, 4 = downstem second, 8 = upstem second
          (if paren-acci 8 0)                             ;8 = parenthesize accidental
          note-id                       ;1 to 12
          0                             ;unknown
          0                             ;unknown
          0                             ;unknown
          (if freeze-acci 2 0)
          ))

(defun entry-flag (note-or-rest groupette break-beam)
  "e.g. (entry-flag 'note 't)"
  (format nil "$~x~x~x~x~x~x~x~x"
          (if note-or-rest 'C '8)
          (if note-or-rest 0 1) ;1 indicates floating rest
          0 ;grace note, articulation
          (if groupette 8 0)
          0 ;hide the entry, beam extension, tie-direction, tie and dot alteration
          (if break-beam 8 0) ;beaming, secondary beam-breaks, freeze stem direction
          0 ;stem alterations, cross-staffing, reverse stem direction for upstems
          0 ;reverse stem direction for all downstems, double stemming, split stemming, stem-direction (1 = up)
          ))

; (defun entry-spec (entry-number prev-entry next-entry duration entry-flag number-of-note-records)
;   "e.g. (entry-spec 2 1 3 1024 '$C0000801 1)"
;   (format nil "~%^eE(~s) ~s ~s ~s ~s ~a ~s ~s "
;           entry-number
;           prev-entry
;           next-entry
;           duration
;           0 ;manual positioning in EVPU's
;           entry-flag
;           128 ;extended-entry-flag: staff expressions, recheck acci's, smart shapes, hide ledger lines, sorted entry
;           number-of-note-records ;12 notes maximum
;           ))

(defun entry-spec (entry-number prev-entry next-entry duration entry-flag note-expression number-of-note-records)
  "e.g. (entry-spec 2 1 3 1024 '$C0000801 nil 1)"
  (format nil "~%^eE(~s) ~s ~s ~s ~s ~a ~s ~s "
          entry-number
          prev-entry
          next-entry
          duration
          0 ;manual positioning in EVPU's
          entry-flag
          (if note-expression 129 128) ;extended-entry-flag
          number-of-note-records ;12 notes maximum
          ))


(defun extended-entry-flag (staff-expression recheck-accis smart-shape hide-ledger-lines sorted-entry)
  "e.g. (extended-entry-flag 'expr nil nil nil nil)"
  (+ (if staff-expression 1 0)
     (if recheck-accis "?" 0)
     (if smart-shape 160 0)
     (if hide-ledger-lines "?" 0)
     (if sorted-entry 128 0)))



; The extended entry flag is a 16 bit value. Although it is a bit flag like the entry flag, it is represented
; in the ETF as a decimal rather than hex value. There is no particular reason for this, and it
; does make the decoding somewhat trickier, but it has been retained to avoid conversion issues.
; #define STAFFEXPRDETAIL0x0001 /* indicates existence of staff expression assignment record */ 
; #define CHECKACCIS 0x0004 /* this bit forces Finale to recompute the ACCIBITs when the frame is loaded. */
; #define SMARTSHAPEDETAIL0x0020 // for smart shape entry attachment #define NOLEGER 0x0040 /* if set, don't draw leger lines for this entry (used for Penderecki-like "highest note possible" stuff) */
; #define ENTRY_IS_SORTED0x0080 /* if set (which it should be most of the time) the notes in the entry are sorted in the normal way, for
; standard notation. If not set, the notes are in some other ordering, which for now would mean percussion notation.*/

; 1
; 2
; 4
; 8
; 16
; 32
; 64
; 128
; 256

; (- 160 128)



;;; obsolete functions:

; (defun entry-flag2 (note-or-rest groupette beam-break)
;   "e.g. (entry-flag2 'note 't 't)"
;   (format nil "$~x~x~x~x~x~x~x~x"
;           (if note-or-rest 'C '8)
;           (if note-or-rest 0 1) ;1 indicates floating rest
;           0 ;grace note, articulation
;           (if groupette 8 0)
;           0 ;hide the entry, beam extension, tie-direction, tie and dot alteration
;           (if beam-break 8 0) ;beaming, secondary beam-breaks, freeze stem direction
;           0 ;stem alterations, cross-staffing, reverse stem direction for upstems
;           0 ;reverse stem direction for all downstems, double stemming, split stemming, stem-direction (1 = up)
;           ))

; (defun make-entry (entry-number prev-entry next-entry entry-data)
;   "e.g. (make-entry 1 0 2 '(q bf4 c4))"
;   (let ((result nil)
;         (pitch-list (cdr entry-data)))
;     (push (entry-spec  entry-number
;                        prev-entry
;                        next-entry
;                        (translate-rhythm (car entry-data))
;                        (entry-flag (if pitch-list 'note) nil nil) 
;                        (length pitch-list))
;           result)
;     (dotimes (i (length pitch-list))
;       (push (note-record (nth i (sort (cdr entry-data) #'< :key #'keynum)) 
;                          (1+ i) 
;                          'show-accidental ;;; this has to change to hide accidentals
;                          nil ;;; this has to change to allow ties: 'start-tie
;                          nil) ;;; this has to change to allow ties: 'end-tie
;             result))
;     (reduce #'(lambda (x y) (concatenate 'string x y)) (reverse result))))

;;; obsolete function:
; (defun make-entries (list-of-events)
;   "e.g. (make-entries '((q bf4 c4) (h. af4 bf3)))"
;   (let (result)
;     (dotimes (i (length list-of-events))
;       (push (make-entry (1+ i) i (+ i 2) (nth i list-of-events)) result))
;     (reduce #'(lambda (x y) (concatenate 'string x y)) (reverse result))))

;;; FRAMES

(defun string+ (&rest list-of-strings)
  (reduce #'(lambda (x y) 
              (concatenate 'string x y)) 
          list-of-strings))

;(get-tag "^FR" *etf-string*)
(defun frame-spec (frame-number starting-entry ending-entry)
  "e.g. (frame-spec 2 7 7)"
  (format nil "~%^FR(~s) ~s ~s 0 0 " 
          frame-number
          starting-entry
          ending-entry))

(defun make-frame-specs (entry-list)
  "e.g. (make-frame-specs '(((h c4) (q c4) (q c4)) ((h c4) (h c4))))"
  (let ((result nil)
        (frame-index 1)
        (entry-index 1))
    (dolist (x (mapcar #'length entry-list))
                (push (frame-spec frame-index entry-index (+ entry-index (1- x))) result)
                (incf frame-index)
                (incf entry-index x))
    (reduce #'string+ (reverse result))))

;(get-tag "^GF" *etf-string*)
(defun frame-ref-spec (staff-number measure-number clef layer1-frame)
;;;this is for Finale 2003
  "e.g. (frame-ref-spec 1 1 'treble 1)"
  (format nil "~%^GF(~s,~s) ~s ~s ~s ~s ~s ~%^GF(~s,~s) ~s ~s ~s ~s ~s"

          staff-number
          measure-number
          (translate-clef clef)
          0  ;unknown
          layer1-frame
          0  ;layer2-frame
          0  ;unknown

          staff-number
          measure-number 
          0  ;layer3-frame
          75 ;an unknown flag - 16-bit bit-vector
          0  ;layer4-frame
          0  ;unknown
          0  ;unknown
          ))

(defun make-frame-ref-specs (clef entry-list)
  "e.g. (make-frame-ref-specs 'treble '(((h c4) (q c4) (q c4)) ((h c4) (h c4))))"
  (let ((result nil))
    (dotimes (i (length entry-list))
      (push (frame-ref-spec 1 (1+ i) clef (1+ i)) result))
    (reduce #'string+ (reverse result))))

;(get-tag "^BC" *etf-string*)

(defun beat-chart-spec (measure-number start-edu horiz-start horiz-end minimum-length barline)
  (format nil "~%^BC(~s) ~s ~s ~s ~s ~s "
          measure-number
          start-edu ;where the element is in the rhythm of the bar
          horiz-start 
          horiz-end
          minimum-length
          (if barline 1 0)))

(defun make-beat-chart (measure-number measure-width duration-list)
  "e.g. (make-beat-chart 1 325 '(3072 1024))"
  (let* ((measure-duration (apply #'+ duration-list))
         (attack-list (cons 0 (reverse (maplist #'(lambda (x) (apply #'+ x)) (reverse duration-list)))))
         (position-list (mapcar #'(lambda (x) (round (* x (/ measure-width measure-duration)))) attack-list)))
    (reduce #'(lambda (x y) (concatenate 'string x y))
            (cons (beat-chart-spec measure-number measure-duration measure-width 40 40 'barline)
                  (mapcar #'(lambda (x y z) (beat-chart-spec measure-number x y z y nil))
                          (butlast attack-list)
                          (butlast position-list)
                          (cdr position-list))))))

(defun beat-chart-stuff (entry-list)
  "e.g. (beat-chart-stuff '(((h c4) (h cs4)) ((h d4) (h ef4))))"
  (let ((durations-in-bars (mapcar #'(lambda (x) 
                                       (mapcar #'(lambda (y) 
                                                   (translate-rhythm (car y))) 
                                               x)) 
                                   entry-list))
        (result nil))
    (dotimes (meas-num (length durations-in-bars) (reduce #'string+ (reverse result)))
      (push (make-beat-chart (1+ meas-num)
                             220 ;measure width in EVPU's
                             (nth meas-num durations-in-bars))
            result))))

; 3.10 Groupettes
; The existence of groupettes are indicated by setting bits in the entry flag and the
; extended entry flag. In the entry flag the start of a groupette is defined by the bit: $C0080000.
; The extended entry flag is set to 8 for each entry that is part of a groupette.
; Detailed information a groupette is stored in the lines that follow the tags ^TN and ^TP in
; the "details" section. These lines do not ordinarily occur in the file, but only exist if groupettes
; are used. The latter line contains the information needed about the duration of the groupette.
; The second number that occurs in the brackets states the entry number that the
; groupette begins on. Five further numbers make up a line of this type. This line indicates how
; many notes of a particular duration make up the groupette and the actual resolves duration of
; the groupette. A quaver triplet on the first entry would be represented by the line:

; ^TP(0,1) 3 512 2 512 3
; This shows that 3 beats of duration 512 make up the groupette and that the actual duration of
; the groupette takes up the space of 2 beats of duration 512.
; Similarly the code to represent a quadruplet of crotchet values is:
; ^TP(0,1) 4 1024 2 1024 3

; (get-tag "^TP" *etf-string*)

; "^TP(0,1) 3 512 2 512 2 
; ^TP(0,1) 0 36 0 16 9745 
; ^TP(0,1) -12 0 -12 0 0 "

(defun tuplet-spec (entry-number 
                    number-of-shown-units 
                    duration-of-shown-units 
                    number-of-real-units 
                    duration-of-real-units)
  "e.g. (tuplet-spec 1 3 512 2 512)" 
  (format nil "~%^TP(~s,~s) ~s ~s ~s ~s ~s ~%^TP(~s,~s) ~s ~s ~s ~s ~s ~%^TP(~s,~s) ~s ~s ~s ~s ~s"
          0 ;always zero
          entry-number
          number-of-shown-units
          duration-of-shown-units
          number-of-real-units
          duration-of-real-units
          2 ;unknown

          0 ;always zero
          entry-number
          0
          36
          0 
          16
          9745 ;9745=number, 9746=ratio, in finale2005 this is either 9873 or 9874
               ;this flag specifies details like whether or not to break the line for the number

          0 ;always zero
          entry-number
          -12
          0
          -12
          0
          0))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;  EXPRESSION DEFINITIONS    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; shape expression: ^DO(1) 1 0 0 0 0 2048

; (defun text-expression-definition-spec (expression-number font-number text-string)
; ; e.g (text-expression-spec 1 2 "crud")
;   (format nil "~%^DT(~s) ~s 0 0 0 0 -16384 ~%^DT(~s) \"~A\""
;           expression-number
;           (+ 3072 font-number)
;           expression-number
;           text-string))

(defun chop-string (string chunk-length &optional (result '()))
  ;; e.g. (chop-string "aaaaaaaaaaaaaaaaaaaaaa" 3)
  (cond ((string-equal string "") (reverse result))
        ((<= (length string) chunk-length)
         (reverse (push string result)))
        (t (chop-string (subseq string chunk-length)
                        chunk-length
                        (push (subseq string 0 chunk-length)
                              result)))))

(defun text-expression-definition-spec (expression-number font-number text-string)
; e.g (text-expression-definition-spec 1 2 "123456789012345678")
  (reduce #'(lambda (x y)
              (concatenate 'string x y))
          (cons (format nil 
                        "~%^DT(~s) ~s 0 0 0 0 -16384 " 
                        expression-number 
                        (+ 3072 font-number))
                (mapcar #'(lambda (string-chunk)
                            (format nil "~%^DT(~s) \"~A\"" expression-number string-chunk))
                        (chop-string text-string 12)))))


(defun make-text-expression-definitions (list-of-strings number-offset)
  ;; e.g. (make-text-expression-definitions '("crap" "help") 28)
  (let (result)
    (dotimes (i (length list-of-strings) result)
      (setf result (concatenate 'string 
                                result 
                                (text-expression-definition-spec (+ i 1 number-offset) ;the expression number
                                                                 2      ;the font = Times New Roman
                                                                 (nth i list-of-strings))))))) ;the string

(defun max-expression-number (etf-string)
  "e.g. (max-expression-number *etf-string*)"
  (let* ((expr-tags (get-tag "^DT" etf-string))
         (last-line-pos (search (format nil "~%") 
                                expr-tags 
                                :from-end t))
         (last-number-end-pos (search (format nil ")") 
                                      expr-tags 
                                      :start2 last-line-pos)))
    (values (read-from-string (subseq expr-tags 
                                      (+ 5 last-line-pos)  
                                      last-number-end-pos)))))

(defun expand-expression-definitions (list-of-strings)
  ;; (reverse (expand-expression-definitions '("crunk" "dunk")))
  (unless (boundp '*etf-string*) (reset-etf))
  (concatenate 'string 
               (get-tag "^DT" *etf-string*)
               (make-text-expression-definitions list-of-strings 
                                                 (max-expression-number *etf-string*))))

(defun expression-number (expr expr-list)
  ;; e.g. (expression-number "crap" '("help" "crap"))
  (1+ (search (list expr) expr-list :test #'equal)))

; (reset-etf)
; (get-tags "^DT")
; (enter-tags (make-text-expression-definitions '("crud" "fun" "dung")))
; (save-etf *enigma-output-filename*)
; (open-finale)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;    EXPRESSION DETAILS      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun expression-details-spec (entry-number expression-number h-pos v-pos)
  "e.g. (expression-details-spec 1 2 0 66)"
  (format nil "~%^ED(~s,~s) ~s ~s ~s ~s ~s "
          0  ;; unknown - why is this always zero?
          entry-number
          expression-number
          h-pos 
          v-pos 
          0
          -32768 ;; position from top/bottom, layer to affect playback, begin playback at assigned note/position in measure
          ))

(defun make-expression-details (entry-number expression-number)
  "e.g. (make-expression-details 1 2)"
  (expression-details-spec entry-number expression-number 0 66))


;; (- 271 16) = 255





;; -32768 = position from top
;; -30720 = position from bottom
;; other stuff to control with this number:
;;   layer to affect playback
;;   begin playback at assigned note / position in measure





; When you add a text expression to an empty file with no libraries, this:

; ^94(65534) 2 360 1800 12 48 18531 
; ^94(65534) 0 1024 0 84 0 16179 
; ^94(65534) 0 0 0 0 0 0 

; changes to this:

; ^94(65534) 2 360 1800 12 48 18531 
; ^94(65534) 0 1024 0 84 0 16180 
; ^94(65534) 0 0 0 0 0 0 

; but omitting this change will not hinder the addition of text expressions, so who cares