;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-

;;; CM-ETF: music notation output from Common Music to Finale format
;;; Copyright (C) 2006 Niklas Kambeitz (niklas.kambeitz@mail.mcgill.ca)
;;;
;;; last modified Time-stamp: <Jul-26-2006 03:23>
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :cm-etf)

(defparameter *template*
  (merge-pathnames "template.ETF" *load-pathname*))

(defparameter *enigma-output-filename*
  (merge-pathnames "cm-etf_OUTPUT.ETF" *load-pathname*))

;; only needed by SBCL on Windows:
(defparameter *path-to-finale* "c:/Program Files/Finale 2006/Finale.exe")

(defparameter *verbose-tag-entry* nil)

(defvar *etf-string*)

(defvar *new-etf*)

(defun load-etf (&optional (filename *template*) (varname '*etf-string*))
  "Loads an enigma file into *etf-string*.
e.g. (load-etf \"empty.etf\")"
  (setf (symbol-value varname) (read-text filename))
  (format t "~&Loaded ~s into ~a" (namestring filename) varname)
  (values))

(cond ((probe-file *template*)
       (load-etf))
      (t (format t "~&~%*** ERROR: Missing ~s. Can't write ETF's~&~%"
                 (namestring *template*))))

(defun reset-etf (&optional (varname '*new-etf*))
  "e.g. (reset-etf)"
  (setf (symbol-value varname) *etf-string*)
;  (format t "~&Reset ~a to *ETF-STRING*" varname)
  (values))

(defun save-etf (&optional (filename *enigma-output-filename*))
  "Write *new-etf* to a file.
e.g. (save-etf)"
  (write-text *new-etf* filename 'string-unix->dos))

(defun open-finale (&optional (filename *enigma-output-filename*))
  "Accepts an optional filename; otherwise opens cm-etf_OUTPUT.ETF with
the default application for ETF's (Windows or Mac)."
  (let ((file (namestring filename)))
    (format t "~%~A~&" "opening enigma file in Finale...")
    #+(and CLISP WIN32) (ext:run-shell-command file :wait nil)
    #+(and SBCL WIN32) (sb-ext:run-program *path-to-finale* (list file) :wait nil)
;    #+(and ACL WIN32) (ext:run-shell-command file :wait nil)
;    #+(and LISPWORKS WIN32) (ext:run-shell-command file :wait nil)

    #+(and CLISP DARWIN) (ext:run-shell-command "open" file :wait nil)
    #+(and CMU DARWIN) (extensions:run-program "open" (list file) :wait nil)
    #+(and SBCL DARWIN) (sb-ext:run-program "open" (list file) :wait nil :search t)
    #+(and OPENMCL DARWIN) (ccl:run-program "open" (list file) :wait nil)
;      
;    #+(and MCL DARWIN) (ext:run-shell-command "open" file :wait nil)
;    #-(or WIN32 DARWIN) nil
    ))

; (sb-ext:run-program (namestring cm-etf::*enigma-output-filename*) nil :wait nil)

(defun section-range (section-name the-string)
  "e.g. (section-range 'details *etf-string*)"
  (let ((start-index (search (format nil "^~(~s~)" section-name) the-string)))
    (if start-index
      (list start-index 
            (apply #'min 
                   (remove-if-not #'(lambda (x) (> x start-index))
                                  (remove nil 
                                          (cons 
                                           (length the-string)
                                           (mapcar #'(lambda (x)
                                                       (search x the-string))
                                                   '("^header" "^others" "^details" 
                                                     "^entries" "^text" "^lyrics"))))))))))

(defun get-section (section-name the-string)
  "e.g. (get-section 'details *etf-string*)"
  (apply #'subseq the-string (section-range section-name the-string)))

(defun replace-section (section-name new-section-string)
  "Destructively replaces section in *new-etf* with new-section-string.
e.g. (replace-section 'text \"a new string\")"
  (let ((section-range (section-range section-name *new-etf*)))
    (setf *new-etf* (concatenate 'string 
                                 (subseq *new-etf* 0 (car section-range))
                                 new-section-string
                                 (subseq *new-etf* (cadr section-range) nil)))))

(defun tag-range (tag the-string)
  "e.g. (tag-range \"^IS\" *etf-string*)"
  (if (search tag the-string)
    (let* ((first-tag-position (search tag the-string))
           (last-tag-position (search tag the-string :from-end 't))
           (end-of-tag (apply #'min 
                              (remove nil (list (search (format nil "~%^") the-string :start2 last-tag-position)
                                                (search (format nil "~%~%") the-string :start2 last-tag-position)
                                                (length the-string))))))
      (list first-tag-position end-of-tag))))

(defun get-tag (tag the-string)
  "Retrieve the lines in the-string beginning with a certain tag.
e.g. (get-tag \"^IS\" *etf-string*)"
  (if (search tag the-string)
    (apply #'subseq the-string (tag-range tag the-string))))

(defun find-next-tag (the-string pos)
  "e.g. (find-next-tag *etf-string* 1000)"
  (if (search (format nil "~%^") the-string :start2 pos)
    (let* ((tag-start-pos (1+ (search (format nil "~%^") the-string :start2 pos)))
           (tag-end-pos (position-if #'(lambda (x) (member x '(#\( #\Space #\Newline)))
                                     the-string :start tag-start-pos)))
      (list (subseq the-string tag-start-pos tag-end-pos) tag-end-pos))))

(defun list-tags (the-string)
 "List all the types of tags present in the-string.
e.g. (list-tags *etf-string*)"
  (do ((result nil (push next-tag result))
       (index 0 (cadr (find-next-tag the-string index)))
       (next-tag nil (car (find-next-tag the-string index))))
      ((null index) (cdr (reverse (remove-duplicates result :test #'equal))))))

(defvar *tag-order* '("^header" "^01" "^02" "^03" "^04" "^05" "^06" "^07" "^08" "^09" "^10" "^11"
                      "^12" "^13" "^14" "^15" "^16" "^17" "^18" "^19" "^20" "^21" "^22" "^23" "^24"
                      "^25" "^26" "^27" "^28" "^29" "^30" "^others" "^&a" "^&c" "^&d" "^&f" "^&r"
                      "^&s" "^&t" "^&u" "^31" "^32" "^33" "^34" "^35" "^36" "^37" "^38" "^39" "^40"
                      "^41" "^42" "^43" "^44" "^45" "^46" "^50" "^51" "^52" "^53" "^55" "^56" "^57"
                      "^58" "^59" "^60" "^61" "^62" "^63" "^64" "^65" "^66" "^67" "^68" "^69" "^70"
                      "^71" "^72" "^73" "^74" "^75" "^76" "^77" "^78" "^79" "^80" "^81" "^82" "^83"
                      "^84" "^85" "^86" "^87" "^88" "^90" "^91" "^92" "^93" "^94" "^95" "^96" "^97"
                      "^98" "^AC" "^BC" "^BK" "^CE" "^DA" "^DC" "^DI" "^DL" "^DO" "^DS" "^DT" "^DY"
                      "^Dc" "^De" "^FB" "^FI" "^FM" "^FN" "^FR" "^FY" "^Fd" "^Fp" "^IK" "^IS" "^IV"
                      "^IX" "^Iu" "^KA" "^LA" "^MN" "^MP" "^MS" "^Mx" "^NS" "^NY" "^Ny" "^OC" "^PD"
                      "^PL" "^PS" "^RS" "^RT" "^SD" "^SS" "^SY" "^Sx" "^Sy" "^TX" "^VS" "^XS" "^dc"
                      "^fI" "^fg" "^ft" "^hp" "^ls" "^ns" "^oc" "^pT" "^pd" "^pg" "^sL" "^sb"
                      "^details" "^BH" "^BL" "^CD" "^CL" "^Cx" "^DE" "^DF" "^DN" "^ED" "^Ex" "^FL"
                      "^GF" "^GL" "^GT" "^IM" "^LP" "^Lx" "^MI" "^NG" "^ST" "^SU" "^TP" "^Ts" "^UE"
                      "^XA" "^XB" "^ac" "^bL" "^ch" "^cn" "^fb" "^mt" "^sB" "^sl" "^ve" "^#v1"
                      "^#v2" "^#v3" "^#v4" "^#v5" "^#v6" "^#v7" "^#v8" "^#v9" "^#v10" "^#c1" "^#c2"
                      "^#c3" "^#c4" "^#c5" "^#c6" "^#c7" "^#c8" "^#c9" "^#c10" "^#s1" "^#s2" "^#s3"
                      "^#s4" "^#s5" "^#s6" "^#s7" "^#s8" "^#s9" "^#s10" "^entries" "^eE" "^text"
                      "^block" "^size" "^font" "^end" "^efx" "^nfx" "^smartshape" "^expression"
                      "^lyrics" "^verse" "^chorus"))

(defun find-insert-point (tag the-string)
  "e.g. (find-insert-point \"^ns\" *etf-string*)"
  (let ((tags-in-string (list-tags the-string)))
    (search (format nil "~%^") the-string
            :start2 (cadr (tag-range (car (remove-if-not #'(lambda (x) (member x tags-in-string :test #'equal))
                                                         (member tag (reverse *tag-order*) :test #'equal)))
                                     the-string)))))

(defun enter-tags (new-string)
  "Puts the new string into *new-etf*, inserting it where 
appropriate, or replacing the old tags as necessary. The
new tag has to start with a #\Newline.
e.g. (enter-tags (make-staves '(bass bass)))"
  (let* ((new-tag-name (car (find-next-tag new-string 0)))
         (insert-range (or (tag-range new-tag-name *new-etf*)
                           (list (find-insert-point new-tag-name *new-etf*)))))
    (if (cadr insert-range) (setq insert-range (list (1- (car insert-range)) (cadr insert-range))))
    (setf *new-etf* (concatenate 'string 
                                 (subseq *new-etf* 0 (car insert-range))
                                 new-string
                                 (subseq *new-etf* (car (last insert-range)))))
    (if *verbose-tag-entry*
        (cond ((cadr insert-range)
               (if *verbose-tag-entry* 
                   (format t "~&~%Replaced ~a tags in *new-etf* with:~&~S" new-tag-name new-string)
                   'replaced-tags))
              (t (if *verbose-tag-entry*
                     (format t "~&~%Inserted tags into *new-etf*:"))
                 'inserted-tags)))))

;(defun enter-tags (new-string)
;  "Puts the new string into *new-etf*, inserting it where 
;appropriate, or replacing the old tags as necessary. The
;new tag has to start with a #\Newline.
;e.g. (enter-tags (make-staff-specs '(bass bass)))"
;  (let* ((new-tag-name (car (find-next-tag new-string 0)))
;         (insert-range (or (tag-range new-tag-name *new-etf*)
;                           (list (find-insert-point new-tag-name *new-etf*)))))
;    (if (cadr insert-range) (setq insert-range (list (1- (car insert-range)) (cadr insert-range))))
;    (setf *new-etf* (concatenate 'string 
;                                 (subseq *new-etf* 0 (car insert-range))
;                                 new-string
;                                 (subseq *new-etf* (car (last insert-range)))))
;    (cond ((cadr insert-range) (format t "~&~%Replaced ~a tags in *new-etf* with:" new-tag-name) (pprint new-string))
;          (t (format t "~&~%Inserted tags into *new-etf*:") (pprint new-string)))))
