;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-

;;; CM-ETF: music notation output in Finale format from Common Music
;;; Copyright (C) 2006 Niklas Kambeitz (niklas.kambeitz@mail.mcgill.ca)
;;;
;;; last modified Time-stamp: <Jul-26-2006 03:22>
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :cm-etf)

(defun find-attacks (&rest rhythm-lists)
  " returns a list of all attacks in a midi sequence or rhythm list
 e.g (find-attacks '((0 1) (3 1)) *counterpoint*)
     (find-attacks #!rhythm-track-0)"
  (let ((result nil))
    (mapcar #'(lambda (x)
                (cond ((equal (type-of x) 'seq) (map-objects #'(lambda (note)
                                                                 (push (object-time note) result))
                                                             x
                                                             :class 'midi))
                      ((equal (type-of x) 'cons) (setf result (mapcar #'first x)))))
            rhythm-lists)
    (sort (remove-duplicates result) #'<)))

(defun find-cut-offs (rhythm)
  " returns a list of all cut-off points in a midi sequence or rhythm list
 e.g (find-cut-offs '((0 1) (3 1)))
     (find-cut-offs #!rhythm-track-0)"
  (let ((result nil))
    (cond  ((equal (type-of rhythm) 'seq) 
            (map-objects #'(lambda (note)
                             (push (+ (object-time note) (midi-duration note)) result))
                         rhythm
                         :class 'midi))
           ((equal (type-of rhythm) 'cons) (dolist (note rhythm) 
                                             (push (+ (first note) (second note)) result))))
    (sort (remove-duplicates result) #'<)))

(defun find-silence (rhythm)
  "returns a list of *attack*-points of all silences in a sequence or rhythm-list.
e.g. (find-silence '((0 1) (3 1)))"
  (sort (set-difference (find-cut-offs rhythm) 
                        (find-attacks rhythm)) 
        #'<))

(defun find-gaps (rhythm)
  "returns a list of rest-events in a sequence or rhythm-list.
e.g. (find-gaps '((0 1) (3 1)))"
  (let ((attacks (find-attacks rhythm)))
    (mapcar #'(lambda (x)
                (list x 
                      (- (find-if #'(lambda (y) (> y x)) attacks)
                         x)))
            (butlast (find-silence rhythm)))))

(defun insert-rests (score-list)
  "e.g. (insert-rests '((0 1) (3 2)))"
  (sort 
   (copy-tree
    (append (find-gaps score-list) score-list))
   #'< :key #'car))

(defun find-all-events (&rest rhythm-lists)
  "returns a list of all attack and cut-off points in a midi sequence(s) and/or rhythm list(s).
e.g.  (find-all-events '((0 1) (3 1)) '((0 2.5) (2.7 1.3)))"
  (sort (remove-duplicates 
         (reduce #'append 
                 (mapcar #'(lambda (x)
                             (append (find-attacks x) (find-cut-offs x)))
                         (copy-list rhythm-lists)))
         :test #'=)
        #'<))

(defun sounding-pitches (time-instant &rest list-of-progs)
  "  Returns a list of sorted pitches sounding at time-instant in sequence(s) and/or event-list(s).
     e.g. (sounding-pitches 6 *mapped-chords* *counterpoint-seq*)"
  (sort
   (copy-list
    (remove-duplicates
     (reduce #'append
             (mapcar #'(lambda (sequence-or-score)
                         (let ((result nil))
                           (cond ((equal (type-of sequence-or-score) 'seq)
                                  (map-objects #'(lambda (x)
                                                   (if (and (<= (object-time x) time-instant)
                                                            (> (+ (object-time x) (midi-duration x)) time-instant))
                                                       (push (midi-keynum x) result)))
                                               (copy-list sequence-or-score)
                                               :class 'midi)
                                  result)
                                 ((equal (type-of sequence-or-score) 'cons)
                                  (mapcar #'(lambda (event)
                                              (if (and (<= (first event) time-instant)
                                                       (> (+ (first event) (second event))  time-instant))
                                                  (setf result (append result (third event)))))
                                          (copy-list sequence-or-score))
                                  result))))
                     (copy-list list-of-progs)))))
   #'<))

(defun score-length (score)
  "e.g. (score-length '((0 2) (2 3)))"
  (apply #'max (mapcar #'(lambda (x)
                           (+ (car x) (cadr x)))
                       score)))
