;;; -*- Syntax: Common-Lisp; Base: 10; Mode: LISP -*-

;;; CM-ETF: music notation output from Common Music to Finale format
;;; Copyright (C) 2006 Niklas Kambeitz (niklas.kambeitz@mail.mcgill.ca)
;;;
;;; last modified Time-stamp: <Jul-26-2006 03:29>
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :cm-etf)

(defvar spelling-table
  '((0 c bs)
    (1 cs df)
    (2 d)
    (3 ef ds)
    (4 e ff)
    (5 f es)
    (6 fs gf)
    (7 g)
    (8 af gs)
    (9 a)
    (10 bf as)
    (11 b cf)))

(defvar pc-spelling-table
  '((bs 110592 69888)
    (c  71936 110592)
    (cs 110592 69888)
    (df 71936 110592)
    (d 110592)
    (ds 110592 69888)
    (ef 71936 110592)
    (e 110592 69888)
    (ff 71936 110592) 
    (es 110592 69888)
    (f 71936 110592)
    (fs 110592 69888)
    (gf 71936 110592)
    (g 110592)
    (gs 110592 69888)
    (af 71936 110592)
    (a 110592)
    (as 110592 69888)
    (bf 71936 110592)
    (b 110592 69888)
    (cf 71936 110592)))

(defun black-notep (pitch)
  "e.g. (black-notep 10)"
  (if (member (mod pitch 12) '(1 3 6 8 10)) 't))

(defun make-spellings (pitch)
  "e.g. (make-spellings 60)"
  (mapcar #'(lambda (x) (transpose (note pitch) x))
          (cdr (assoc (note-name pitch) pc-spelling-table))))

(defun spell-pitch (pitch)
  "e.g. (spell-pitch 'c)"
    (if (or (and (numberp pitch) (< pitch 12))
            (null (octave-number pitch)))
      (mapcar #'note-name (make-spellings pitch))
      (make-spellings pitch)))

(defun possible-spellings (pitches)
  "e.g. (possible-spellings '(59 65 72)) (possible-spellings 60) (possible-spellings 'c)"
  (if (listp pitches) 
    (cartesian (mapcar #'spell-pitch pitches)) 
    (spell-pitch pitches)))

;(interval-quality 119558)
;(interval-quality '(c5 e5))
;(interval-quality (interval :perfect 4))

(defun interval-penalty (interval-or-note-list)
  "e.g. (interval-penalty (interval :perfect 4)) (interval-penalty '(ff5 bs5))"
  (cond ((numberp interval-or-note-list) 
         (cond ((member (interval-quality interval-or-note-list) '(perfect major minor)) 0)
               ((member interval-or-note-list '(119558 103430)) 0) ; Tritones
               ((member (interval-quality interval-or-note-list) '(augmented diminished)) 1)
               ((member (interval-quality interval-or-note-list) '(doubly-diminished doubly-augmented)) 3)))
        ((listp interval-or-note-list) 
         (if (subsetp '(bs ff) (mapcar #'note-name interval-or-note-list)) 100
             ;; CM's INTERVAL can't handle triply diminished or
             ;; augmented intervale, so they get a huge penalty.
             (reduce #'+ 
                     (mapcar #'(lambda (x)
                                 (apply #'interval x)) 
                             (combinations interval-or-note-list 2))
                     :key #'interval-penalty)))))

(defun note-penalty (note-or-chord)
  "e.g. (note-penalty '(ff bs gf))"
  (cond ((atom note-or-chord)
         (cond ((member (note-name note-or-chord) '(c cs d ef e f fs g gs af a bf b)) 0) ;preferred-spellings
               ((member (note-name note-or-chord) '(df ds gf as)) 0.01) ;less-preferred spellings
               ((member (note-name note-or-chord) '(bs cf es ff)) 4) ;ugly spellings
               ((member (note-name note-or-chord) '(dff css eff dss gff fss aff gss bff ass)) 4) ;double-accidentals
               (t (error "note-name not listed in NOTE-PENALTY"))))
        ((listp note-or-chord)
         (reduce #'+ (mapcar #'note-penalty note-or-chord)))))

(defun spelling-penalty (note-list)
  "e.g. (spelling-penalty '(c e))"
  (+ (interval-penalty note-list) (note-penalty note-list)))

;(interval-penalty '(c5 e5))
;(note-penalty '(c e))

(defun spell (pitch-or-pitch-list)
  "SPELL tries to correctly spell a list of midi pitch numbers.
e.g. (spell '(58 59 63))"
  (cond ((listp pitch-or-pitch-list) 
         (car (sort (possible-spellings pitch-or-pitch-list) #'< :key #'spelling-penalty)))
        ((or (and (numberp pitch-or-pitch-list) (< pitch-or-pitch-list 12)) 
             (null (octave-number pitch-or-pitch-list))) 
         (note-name pitch-or-pitch-list))
        (t (note pitch-or-pitch-list))))

;(defun spell-with-ref (pitches ref-pitches)
;  "e.g. (spell-with-ref '(10) '(11))"
;  (let* ((pc-pool (find-pcs (append pitches ref-pitches)))
;         (spelling-alist (mapcar #'list pc-pool (spell pc-pool))))
;    (mapcar #'(lambda (x)
;                (cadr (assoc (pitch-class x) spelling-alist)))
;            pitches)))
